from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.magnetic_structures.srw_bending_magnet import SRWBendingMagnet
from wofrysrw.storage_ring.srw_electron_beam import SRWElectronBeam

from syned.storage_ring.magnetic_structures.bending_magnet import BendingMagnet

class SRWBendingMagnetLightSource(SRWLightSource):

    def __init__(self,
                 name="Undefined",
                 electron_beam=SRWElectronBeam(),
                 bending_magnet_magnetic_structure=SRWBendingMagnet()):


        super().__init__(name,
                         electron_beam=electron_beam,
                         magnetic_structure=bending_magnet_magnetic_structure)



import numpy

from oasys_srw.srwlib import SRWLPartBeam

from syned.storage_ring.electron_beam import ElectronBeam

from wofrysrw.srw_object import SRWObject

class SRWElectronBeamDecorator():

    def to_SRWLPartBeam(self):
        pass

    @classmethod
    def from_SRWLPartBeam(cls, srw_part_beam, number_of_bunches = 400):
        pass

class SRWElectronBeamGeometricalProperties(object):
    def __init__(self,
                 electron_beam_size_h = 0.0,
                 electron_beam_divergence_h = 0.0,
                 electron_beam_size_v = 0.0,
                 electron_beam_divergence_v = 0.0):
        self._electron_beam_size_h = electron_beam_size_h
        self._electron_beam_divergence_h = electron_beam_divergence_h
        self._electron_beam_size_v = electron_beam_size_v
        self._electron_beam_divergence_v = electron_beam_divergence_v

    def to_info(self):
        info = 'Electron beam : \n'
        info += '   RMS size H/V [um]: '+ repr(self._electron_beam_size_h*1e6) + '  /  ' + repr(self._electron_beam_size_v*1e6) + '\n'
        info += '   RMS divergence H/V [urad]: '+ repr(self._electron_beam_divergence_h*1e6) + '  /  ' + repr(self._electron_beam_divergence_v*1e6) + '\n\n'

        return info

class SRWElectronBeam(ElectronBeam, SRWElectronBeamDecorator, SRWObject):
    def __init__(self,
                 energy_in_GeV = 1.0,
                 energy_spread = 0.0,
                 current = 0.1,
                 number_of_bunches = 400,
                 moment_x=0.0,
                 moment_xp=0.0,
                 moment_y=0.0,
                 moment_yp=0.0,
                 moment_z=0.0,
                 moment_xx=0.0,
                 moment_xxp=0.0,
                 moment_xpxp=0.0,
                 moment_yy=0.0,
                 moment_yyp=0.0,
                 moment_ypyp=0.0):
        super().__init__(energy_in_GeV,
                         energy_spread,
                         current,
                         number_of_bunches,
                         moment_xx,
                         moment_xxp,
                         moment_xpxp,
                         moment_yy,
                         moment_yyp,
                         moment_ypyp)

        self._moment_x = moment_x
        self._moment_xp = moment_xp
        self._moment_y = moment_y
        self._moment_yp = moment_yp
        self._moment_z = moment_z

    def set_moments_from_twiss(self, horizontal_emittance = 0.0,
                                     horizontal_beta = 0.0,
                                     horizontal_alpha = 0.0,
                                     horizontal_eta = 0.0,
                                     horizontal_etap   = 0.0,
                                     vertical_emittance=0.0,
                                     vertical_beta     = 0.0,
                                     vertical_alpha    = 0.0,
                                     vertical_eta = 0.0,
                                     vertical_etap = 0.0):
        
        srw_electron_beam = self.to_SRWLPartBeam()
        srw_electron_beam.from_Twiss(_Iavg    =self._current,
                                     _e       =self._energy_in_GeV,
                                     _sig_e   =self._energy_spread,
                                     _emit_x  =horizontal_emittance,
                                     _beta_x  =horizontal_beta,
                                     _alpha_x =horizontal_alpha,
                                     _eta_x   =horizontal_eta,
                                     _eta_x_pr=horizontal_etap,
                                     _emit_y  =vertical_emittance,
                                     _beta_y  =vertical_beta,
                                     _alpha_y =vertical_alpha,
                                     _eta_y   =vertical_eta,
                                     _eta_y_pr=vertical_etap)
                                     
        srw_electron_beam = SRWElectronBeam.from_SRWLPartBeam(srw_electron_beam, self._number_of_bunches)

        self._energy_in_GeV       = srw_electron_beam._energy_in_GeV
        self._energy_spread       = srw_electron_beam._energy_spread
        self._current             = srw_electron_beam._current
        self._number_of_bunches   = srw_electron_beam._number_of_bunches
        self._moment_xx           = srw_electron_beam._moment_xx
        self._moment_xxp          = srw_electron_beam._moment_xxp
        self._moment_xpxp         = srw_electron_beam._moment_xpxp
        self._moment_yy           = srw_electron_beam._moment_yy
        self._moment_yyp          = srw_electron_beam._moment_yyp
        self._moment_ypyp         = srw_electron_beam._moment_ypyp
        self._moment_x            = srw_electron_beam._moment_x
        self._moment_xp           = srw_electron_beam._moment_xp
        self._moment_y            = srw_electron_beam._moment_y
        self._moment_yp           = srw_electron_beam._moment_yp
        self._moment_z            = srw_electron_beam._moment_z

    def set_moments_from_electron_beam_geometrical_properties(self, electron_beam_geometrical_properties = SRWElectronBeamGeometricalProperties()):
        self.set_sigmas_all(sigma_x=electron_beam_geometrical_properties._electron_beam_size_h,
                            sigma_xp=electron_beam_geometrical_properties._electron_beam_divergence_h,
                            sigma_y=electron_beam_geometrical_properties._electron_beam_size_v,
                            sigma_yp=electron_beam_geometrical_properties._electron_beam_divergence_v)

    def get_electron_beam_geometrical_properties(self):
        x, xp, y, yp = self.get_sigmas_all()

        return SRWElectronBeamGeometricalProperties(electron_beam_size_h=x,
                                                    electron_beam_divergence_h=xp,
                                                    electron_beam_size_v=y,
                                                    electron_beam_divergence_v=yp)
    def to_SRWLPartBeam(self):
        srw_electron_beam = SRWLPartBeam()
        srw_electron_beam.Iavg           = self._current

        srw_electron_beam.partStatMom1.x = self._moment_x
        srw_electron_beam.partStatMom1.y = self._moment_y
        srw_electron_beam.partStatMom1.z = self._moment_z
        srw_electron_beam.partStatMom1.xp = self._moment_xp
        srw_electron_beam.partStatMom1.yp = self._moment_yp
        srw_electron_beam.partStatMom1.gamma = self.gamma()

        #2nd order statistical moments:
        srw_electron_beam.arStatMom2[0] = self._moment_xx   # <(x-x0)^2> [m^2]
        srw_electron_beam.arStatMom2[1] = self._moment_xxp  # <(x-x0)*(x'-x'0)> [m]
        srw_electron_beam.arStatMom2[2] = self._moment_xpxp # <(x'-x'0)^2>
        srw_electron_beam.arStatMom2[3] = self._moment_yy   #<(y-y0)^2>
        srw_electron_beam.arStatMom2[4] = self._moment_yyp  #<(y-y0)*(y'-y'0)> [m]
        srw_electron_beam.arStatMom2[5] = self._moment_ypyp #<(y'-y'0)^2>
        srw_electron_beam.arStatMom2[10] = self._energy_spread**2 #<(E-E0)^2>/E0^2
                
        return srw_electron_beam

    @classmethod
    def from_SRWLPartBeam(cls, srw_part_beam, number_of_bunches = 400):
        srw_electron_beam = SRWElectronBeam(energy_spread = numpy.sqrt(srw_part_beam.arStatMom2[10]),
                                            current = srw_part_beam.Iavg,
                                            number_of_bunches = number_of_bunches,
                                            moment_x=srw_part_beam.partStatMom1.x,
                                            moment_y=srw_part_beam.partStatMom1.y,
                                            moment_z=srw_part_beam.partStatMom1.z,
                                            moment_xp=srw_part_beam.partStatMom1.xp,
                                            moment_yp=srw_part_beam.partStatMom1.yp,
                                            moment_xx  = srw_part_beam.arStatMom2[0],
                                            moment_xxp = srw_part_beam.arStatMom2[1],
                                            moment_xpxp= srw_part_beam.arStatMom2[2],
                                            moment_yy  = srw_part_beam.arStatMom2[3],
                                            moment_yyp = srw_part_beam.arStatMom2[4],
                                            moment_ypyp= srw_part_beam.arStatMom2[5])

        srw_electron_beam.set_energy_from_gamma(srw_part_beam.partStatMom1.gamma)

        return srw_electron_beam

    def to_python_code(self, data=None):
        text_code  = "part_beam = SRWLPartBeam()" + "\n"
        text_code += "part_beam.Iavg               = " + str(self._current) + "\n"
        text_code += "part_beam.partStatMom1.x     = " + str(self._moment_x) + "\n"
        text_code += "part_beam.partStatMom1.y     = " + str(self._moment_y) + "\n"
        text_code += "part_beam.partStatMom1.z     = " + str(self._moment_z) + "\n"
        text_code += "part_beam.partStatMom1.xp    = " + str(self._moment_xp) + "\n"
        text_code += "part_beam.partStatMom1.yp    = " + str(self._moment_yp) + "\n"
        text_code += "part_beam.partStatMom1.gamma = " + str(self.gamma()) + "\n"
        text_code += "part_beam.arStatMom2[0]      = " + str(self._moment_xx) + "\n"
        text_code += "part_beam.arStatMom2[1]      = " + str(self._moment_xxp) + "\n"
        text_code += "part_beam.arStatMom2[2]      = " + str(self._moment_xpxp) + "\n"
        text_code += "part_beam.arStatMom2[3]      = " + str(self._moment_yy) + "\n"
        text_code += "part_beam.arStatMom2[4]      = " + str(self._moment_yyp) + "\n"
        text_code += "part_beam.arStatMom2[5]      = " + str(self._moment_ypyp) + "\n"
        text_code += "part_beam.arStatMom2[10]     = " + str(self._energy_spread**2) + "\n"

        return text_code



import numpy

from scipy.interpolate import RectBivariateSpline
from oasys.util.oasys_util import read_surface_file

def add_thickness_error_to_thickness_profile(thickness_profile,
                                             thickness_error_profile_file,
                                             scaling_factor=1.0,
                                             x_range=[-1e-3, 1e-3],
                                             y_range=[-1e-3, 1e-3],
                                             n_points_x=100,
                                             n_points_y=100):
    xx, yy, zz = read_surface_file(thickness_error_profile_file)
    interpolator = RectBivariateSpline(xx, yy, (zz.T) * scaling_factor)

    xx_t = numpy.linspace(x_range[0], x_range[1], n_points_x)
    yy_t = numpy.linspace(y_range[0], y_range[1], n_points_y)

    for i in range(n_points_x):
        for j in range(n_points_y):
            thickness_profile[i, j] += interpolator.ev(xx_t[i], yy_t[j])

def add_thickness_error_transmission(srwlopt_t, n_points_x, n_points_y,
                                     error_transmission_amplitudes, error_transmission_optical_path_difference):
    ofst = 0
    for iy in range(n_points_y):
        for ix in range(n_points_x):
            srwlopt_t.arTr[ofst] *= error_transmission_amplitudes[ix, iy]
            srwlopt_t.arTr[ofst + 1] += error_transmission_optical_path_difference[ix, iy]
            ofst += 2

def get_transmission_amplitudes(thickness_profile, attenuation_length):
    return numpy.exp(-0.5*thickness_profile/attenuation_length)

def get_transmission_optical_path_difference(thickness_profile, delta):
    return -delta*thickness_profile



import numpy
from oasys_srw.srwlib import array as srw_array

def numpyComplexArrayToSRWArray(numpy_array):
    """
    Converts a numpy.array to an array usable by SRW.
    :param numpy_array: a 2D numpy array
    :return: a 2D complex SRW array
    """
    elements_size = numpy_array.size

    r_horizontal_field = numpy_array[:, :].real.transpose().flatten().astype(numpy.float)
    i_horizontal_field = numpy_array[:, :].imag.transpose().flatten().astype(numpy.float)

    tmp = numpy.zeros(elements_size * 2, dtype=numpy.float32)
    for i in range(elements_size):
        tmp[2*i] = r_horizontal_field[i]
        tmp[2*i+1] = i_horizontal_field[i]

    return srw_array('f', tmp)

def numpyArraysToSRWArray(numpy_array_re, numpy_array_im):
    """
    Converts a numpy.array to an array usable by SRW.
    :param numpy_array: a 2D numpy array
    :return: a 2D complex SRW array
    """
    elements_size = numpy_array_re.size

    r_horizontal_field = numpy_array_re[:, :].transpose().flatten().astype(numpy.float)
    i_horizontal_field = numpy_array_im[:, :].transpose().flatten().astype(numpy.float)

    tmp = numpy.zeros(elements_size * 2, dtype=numpy.float32)
    for i in range(elements_size):
        tmp[2*i]   = r_horizontal_field[i]
        tmp[2*i+1] = i_horizontal_field[i]

    return srw_array('f', tmp)

def __reshape(numpy_array, dim_x, dim_y, number_energies, polarized=True):
    if polarized: numpy_array = numpy_array.reshape((dim_y, dim_x, number_energies, 1))
    else:         numpy_array.reshape((dim_y, dim_x, number_energies))
    numpy_array = numpy_array.swapaxes(0, 2)

    return numpy_array.copy()

def SRWArrayToNumpyComplexArray(srw_array, dim_x, dim_y, number_energies, polarized=True):
    """
    Converts a SRW array to a numpy.array.
    :param srw_array: SRW array
    :param dim_x: size of horizontal dimension
    :param dim_y: size of vertical dimension
    :param number_energies: Size of energy dimension
    :return: 4D numpy array: [energy, horizontal, vertical, polarisation={0:horizontal, 1: vertical}]
    """
    re = numpy.array(srw_array[::2], dtype=numpy.float)
    im = numpy.array(srw_array[1::2], dtype=numpy.float)

    return __reshape(re + 1j * im, dim_x, dim_y, number_energies, polarized)

def SRWArrayToNumpyArrays(srw_array, dim_x, dim_y, number_energies, polarized=True):
    """
    Converts a SRW array to a numpy.array.
    :param srw_array: SRW array
    :param dim_x: size of horizontal dimension
    :param dim_y: size of vertical dimension
    :param number_energies: Size of energy dimension
    :return: 4D numpy array: [energy, horizontal, vertical, polarisation={0:horizontal, 1: vertical}]
    """
    re = numpy.array(srw_array[::2], dtype=numpy.float)
    im = numpy.array(srw_array[1::2], dtype=numpy.float)

    return __reshape(re, dim_x, dim_y, number_energies, polarized), \
           __reshape(im, dim_x, dim_y, number_energies, polarized)




from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.magnetic_structures.srw_3d_magnetic_structure import SRW3DMagneticStructure
from wofrysrw.storage_ring.srw_electron_beam import SRWElectronBeam

class SRW3DLightSource(SRWLightSource):

    def __init__(self,
                 name="Undefined",
                 electron_beam=SRWElectronBeam(),
                 magnet_magnetic_structure=SRW3DMagneticStructure()):

        super().__init__(name,
                         electron_beam=electron_beam,
                         magnetic_structure=magnet_magnetic_structure)



import numpy

from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.light_sources.srw_wiggler_light_source import SRWWigglerLightSource
from wofrysrw.storage_ring.srw_light_source import SourceWavefrontParameters, WavefrontPrecisionParameters

from srxraylib.plot.gol import plot, plot_contour, plot_surface

wiggler = SRWWigglerLightSource("XRD1",
                                electron_energy_in_GeV=2.0,
                                electron_energy_spread=0.0007,
                                ring_current=0.4,
                                electron_beam_size_h=0.05545e-3,
                                electron_beam_size_v=2.784e-6,
                                emittance=0.2525e-9,
                                coupling_costant=0.01,
                                K_vertical=20.9155,
                                period_length=0.064,
                                number_of_periods=24.5)

print(wiggler.get_electron_beam().get_electron_beam_geometrical_properties().to_info())


wf_parameters = SourceWavefrontParameters(photon_energy_min = 13000,
                                          photon_energy_max = 13000,
                                          photon_energy_points=1,
                                          h_slit_gap = 10e-3,
                                          v_slit_gap = 2e-3,
                                          h_slit_points=500,
                                          v_slit_points=50,
                                          distance = 10.0,
                                          wavefront_precision_parameters=WavefrontPrecisionParameters(sr_method=2,
                                                                                                      relative_precision=0.01,
                                                                                                      number_of_points_for_trajectory_calculation=20000,
                                                                                                      sampling_factor_for_adjusting_nx_ny=-1))

e, h, v, i = wiggler.get_intensity(source_wavefront_parameters=wf_parameters, multi_electron=False)


plot_contour(i[int(int(e.size/2))],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("XRD1",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]",plot_points=0,
             contour_levels=numpy.linspace(0, i.max(), 20), cmap=None, cbar=1,cbar_title="Flux ",show=1)

plot_surface(i[int(e.size/2)],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("XRD1",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]", show=1)


from oasys_srw.srwlib import SRWLMagFldH, SRWLMagFldU

from syned.storage_ring.magnetic_structures.undulator import Undulator
from wofrysrw.storage_ring.srw_magnetic_structure import SRWMagneticStructure

class SRWUndulator(Undulator, SRWMagneticStructure):

    def __init__(self,
                 horizontal_central_position = 0.0,
                 vertical_central_position = 0.0,
                 longitudinal_central_position = 0.0,
                 K_vertical = 0.0,
                 K_horizontal = 0.0,
                 period_length = 0.0,
                 number_of_periods = 1,
                 initial_phase_vertical = 0.0, # initial phase [rad]
                 initial_phase_horizontal = 0.0, # initial phase [rad]
                 symmetry_vs_longitudinal_position_vertical = -1, # symmetry vs longitudinal position 1 - symmetric (B ~ cos(2*Pi*n*z/per + ph)) , -1 - anti-symmetric (B ~ sin(2*Pi*n*z/per + ph))
                 symmetry_vs_longitudinal_position_horizontal = 1, # symmetry vs longitudinal position 1 - symmetric (B ~ cos(2*Pi*n*z/per + ph)) , -1 - anti-symmetric (B ~ sin(2*Pi*n*z/per + ph))
                 coefficient_for_transverse_dependence_vertical = 1.0, # coefficient for transverse depenednce B*cosh(2*Pi*n*a*y/per)*cos(2*Pi*n*z/per + ph)
                 coefficient_for_transverse_dependence_horizontal = 1.0 # coefficient for transverse depenednce B*cosh(2*Pi*n*a*y/per)*cos(2*Pi*n*z/per + ph)
                ):
        Undulator.__init__(self, K_vertical, K_horizontal, period_length, number_of_periods)
        SRWMagneticStructure.__init__(self, horizontal_central_position, vertical_central_position, longitudinal_central_position)

        self.initial_phase_vertical = initial_phase_vertical
        self.initial_phase_horizontal = initial_phase_horizontal
        self.symmetry_vs_longitudinal_position_vertical = symmetry_vs_longitudinal_position_vertical
        self.symmetry_vs_longitudinal_position_horizontal = symmetry_vs_longitudinal_position_horizontal
        self.coefficient_for_transverse_dependence_vertical = coefficient_for_transverse_dependence_vertical
        self.coefficient_for_transverse_dependence_horizontal = coefficient_for_transverse_dependence_horizontal

    def get_SRWMagneticStructure(self):
        magnetic_fields = []

        if self._K_vertical > 0.0:
            magnetic_fields.append(SRWLMagFldH(1,
                                               'v',
                                               _B=self.magnetic_field_vertical(),
                                               _ph=self.initial_phase_vertical,
                                               _s=self.symmetry_vs_longitudinal_position_vertical,
                                               _a=self.coefficient_for_transverse_dependence_vertical))

        if self._K_horizontal > 0.0:
            magnetic_fields.append(SRWLMagFldH(1,
                                               'h',
                                               _B=self.magnetic_field_horizontal(),
                                               _ph=self.initial_phase_horizontal,
                                               _s=self.symmetry_vs_longitudinal_position_horizontal,
                                               _a=self.coefficient_for_transverse_dependence_horizontal))

        return SRWLMagFldU(_arHarm=magnetic_fields,
                           _per=self._period_length,
                           _nPer=self._number_of_periods)

    def to_python_code_aux(self):
        text_code = "magnetic_fields = []" + "\n"

        if self._K_vertical > 0.0:
            text_code += "magnetic_fields.append(SRWLMagFldH(1, 'v', " + "\n"
            text_code += "                                   _B=" + str(self.magnetic_field_vertical()) + ", " + "\n"
            text_code += "                                   _ph=" + str(self.initial_phase_vertical) + ", " + "\n"
            text_code += "                                   _s=" + str(self.symmetry_vs_longitudinal_position_vertical) + ", " + "\n"
            text_code += "                                   _a=" + str(self.coefficient_for_transverse_dependence_vertical) + "))" + "\n"

        if self._K_horizontal > 0.0:
            text_code += "magnetic_fields.append(SRWLMagFldH(1, 'h', " + "\n"
            text_code += "                                   _B=" + str(self.magnetic_field_horizontal()) + ", " + "\n"
            text_code += "                                   _ph=" + str(self.initial_phase_horizontal) + ", " + "\n"
            text_code += "                                   _s=" + str(self.symmetry_vs_longitudinal_position_horizontal) + ", " + "\n"
            text_code += "                                   _a=" + str(self.coefficient_for_transverse_dependence_horizontal) + "))" + "\n"

        text_code += "magnetic_structure = SRWLMagFldU(_arHarm=magnetic_fields, _per=" + str(self._period_length) + ", _nPer=" + str(self._number_of_periods) + ")" + "\n"

        return text_code
